'''
    File: voting_methods_registry.py
    Author: Wes Holliday (wesholliday@berkeley.edu) and Eric Pacuit (epacuit@umd.edu)
    Date: April 29, 2024
    
'''

from pref_voting.voting_method import VotingMethod
import inspect

class VotingMethodRegistry:
    def __init__(self):
        self.methods = {}

    def register(self, 
                 method, 
                 name, 
                 method_type, 
                 file_location, 
                 election_types, 
                 properties):
        self.methods[name] = {
            "method": method,
            "method_type": method_type,
            "file_location": file_location,
            "election_types": election_types,
            "properties": properties
        }

    def discover_methods(self, module):
        """Discovers and registers all VotingMethod instances in the given module."""

        fname_to_method_type = {
            "scoring_methods.py": "Scoring Rule",
            "iterative_methods.py": "Iterative Method",
            "c1_methods.py": "C1 Method",
            "margin_based_methods.py": "Margin Based Method",
            "other_methods.py": "Other Method",
            "combined_methods.py": "Combined Method",
        }
        for name, obj in inspect.getmembers(module, lambda member: isinstance(member, VotingMethod)):
            if hasattr(obj, 'name') and hasattr(obj, 'properties'):

                if getattr(obj, 'skip_registration', False):
                    continue

                self.register(obj, 
                              obj.name, 
                              fname_to_method_type[module.__file__.split("/")[-1]],
                              module.__file__.split("/")[-1], 
                              getattr(obj, 'input_types', []), 
                              getattr(obj, 'properties', {}))

    def list_methods(self):
        return list(self.methods.keys())

    def display_methods(self):
        for name, details in self.methods.items():
            print(f"{name} ({details['method_type']})")
            print(f"Satisfied properties: {details['properties'].satisfied() if details['properties'] is not None else 'N/A'}")
            print(f"Violated properties: {details['properties'].violated() if details['properties'] is not None else 'N/A'}")
            print()

    def get_methods(self, method_type): 
        return [self.methods[vm]['method'] for vm in self.methods if self.methods[vm]['method_type'] == method_type]

    def filter(self, 
               satisfies=None, 
               violates=None, 
               unknown=None, 
               election_types=None):
        
        if satisfies is None:
            satisfies = []
        if violates is None:
            violates = []
        if unknown is None:
            unknown = []
        if election_types is None:
            election_types = []

        found_methods = []
        for name, method_info in self.methods.items():
            properties = method_info['properties']
            method_election_types = method_info['election_types']

            # Check if the method satisfies the required properties
            if not all(getattr(properties, prop, None) is True for prop in satisfies):
                continue

            # Check if the method violates the required properties
            if not all(getattr(properties, prop, None) is False for prop in violates):
                continue

            # Check for unknown properties
            if not all(hasattr(properties, prop) and getattr(properties, prop, None) is None for prop in unknown):
                continue

            # Check if the method supports all required election types
            if not all(et in method_election_types for et in election_types):
                continue

            # If all conditions are met, add to results
            found_methods.append(method_info['method'])

        return found_methods

    def method_type(self, method_name):
        return self.methods[method_name]['method_type']
    def file_location(self, method_name):
        return self.methods[method_name]['file_location']
    
    def __len__(self):
        return len(self.methods)
    
    def __iter__(self):
        self._iter = iter(method_details['method'] for method_details in self.methods.values())
        return self

    def __next__(self):
        return next(self._iter)

voting_methods = VotingMethodRegistry()

import pref_voting.scoring_methods 
voting_methods.discover_methods(pref_voting.scoring_methods)
import pref_voting.iterative_methods 
voting_methods.discover_methods(pref_voting.iterative_methods)
import pref_voting.c1_methods 
voting_methods.discover_methods(pref_voting.c1_methods)
import pref_voting.margin_based_methods 
voting_methods.discover_methods(pref_voting.margin_based_methods)
import pref_voting.combined_methods 
voting_methods.discover_methods(pref_voting.combined_methods)
import pref_voting.other_methods 
voting_methods.discover_methods(pref_voting.other_methods)



'''
    File: num_evaluation_method.py
    Author: Wes Holliday (wesholliday@berkeley.edu) and Eric Pacuit (epacuit@umd.edu)
    Date: June 2, 2024
    
    The NumEvaluationMethod class and helper functions for numerical evaluation methods.
'''

import functools
import numpy as np
from pref_voting.mappings import Utility
from pref_voting.rankings import Ranking

class NumEvaluationMethod(object): 
    """
    A class to add functionality to numerical evaluation methods 

    Args:
        nem (function): An implementation of a numerical evaluation method. The function should accept any type of profile, and a keyword parameter ``curr_cands`` to find the numerical evaluation after restricting to ``curr_cands``. 
        name (string): The human-readable name of the numerical evaluation function.
    
    Returns: 
        A utility function (Utility) that represents the numerical evaluation of each candidate.

    """
    def __init__(self, nem, name = None): 
        
        self.nem = nem
        self.name = name
        functools.update_wrapper(self, nem)   

    def __call__(self, edata, curr_cands = None, **kwargs):

        if (curr_cands is not None and len(curr_cands) == 0) or len(edata.candidates) == 0: 
            return Utility({})
        return self.nem(edata, curr_cands = curr_cands, **kwargs)
        
    def ranking(self, edata, curr_cands = None, **kwargs): 
        """
        Return the ranking generated by the numerical evaluation. 
        """

        if (curr_cands is not None and len(curr_cands) == 0) or len(edata.candidates) == 0: 
            return Ranking({})
        ev =  self.nem(edata, curr_cands = curr_cands, **kwargs)
        return ev.ranking()
        
    def display(self, edata, curr_cands = None, cmap = None, **kwargs): 
        """
        Display the winning set of candidates.
        """
 
        cmap = cmap if cmap is not None else edata.cmap

        ev = self.__call__(edata, curr_cands = curr_cands, **kwargs)

        if ev is None:  # some voting methods may return None if, for instance, it is taking long to compute the winner.
            print(f"{self.name} numerical evaluation is not available")
        else: 
            w_str = f"{self.name} evaluation is " 
            print(w_str + str(ev))
        
    def set_name(self, new_name):
        """Set the name of the social welfare function."""

        self.name = new_name

    def __str__(self): 
        return f"{self.name}"

def nem(name = None):
    """
    A decorator used when creating a numerical evaluation method. 
    """
    def wrapper(f):
        return NumEvaluationMethod(f, name=name)
    return wrapper



"""
    File: swf_axioms.py
    Author: Wesley H. Holliday (wesholliday@berkeley.edu) and Eric Pacuit (epacuit@umd.edu)
    Date: April 29, 2024
    
    SWF axioms 
"""

from pref_voting.axiom import Axiom
from pref_voting.axiom_helpers import *
from itertools import permutations
from pref_voting.helper import weak_orders

def has_pareto_ranking_violation(prof,swf,verbose=False, strong_Pareto=False):

    """Returns True if all voters rank x above y in prof, but the SWF does not rank x above y. 
    
    If verbose is True, prints the profile and the violation.

    If strong_Pareto is True, then a violation occurs when all voters weakly prefer x to y, some voter strictly prefers x to y, but the SWF does not rank x above y.
    
    Args:
        prof: a Profile or ProfileWithTies
        swf (SocialWelfareFunction): An SWF to test.
        verbose (bool, default=False): If a violation is found, display the violation.

    Returns:
        Result of the test (bool): Returns True if there is a violation and False otherwise.
    
    """
    social_ranking = swf(prof)

    for x in prof.candidates: 
        for y in prof.candidates: 
            if not social_ranking.extended_strict_pref(x,y): 
                if (strong_Pareto == False and prof.support(x,y)==prof.num_voters) or (strong_Pareto == True and prof.support(x,y)> 0 and prof.support(y,x)==0):
                    if verbose:  
                        print(f"Pareto ranking violation by {swf}:")
                        prof.display()
                        print(prof.description())
                        swf.display(prof)
                        print(f"Candidate {x} Pareto dominates {y} but is not ranked above {y} by {swf}.")
                        print()
                    return True
    return False

def find_all_pareto_ranking_violations(prof,swf,verbose=False, strong_Pareto=False):

    """Returns a list of all pairs of candidates for which the SWF violates the Pareto ranking axiom. 
    
    If verbose is True, prints the profile and the violation.

    If strong_Pareto is True, then a violation occurs when all voters weakly prefer x to y, some voter strictly prefers x to y, but the SWF does not rank x above y.
    
    Args:
        prof: a Profile or ProfileWithTies
        swf (SocialWelfareFunction): An SWF to test.
        verbose (bool, default=False): If a violation is found, display the violation.

    Returns:
        List of violations (list): List of all pairs of candidates for which there is a violation.
    
    """
    social_ranking = swf(prof)

    violations = []

    for x in prof.candidates: 
        for y in prof.candidates: 
            if not social_ranking.extended_strict_pref(x,y): 
                if (strong_Pareto == False and prof.support(x,y)==prof.num_voters) or (strong_Pareto == True and prof.support(x,y)> 0 and prof.support(y,x)==0):
                    if verbose:  
                        print(f"Pareto ranking violation by {swf}:")
                        prof.display()
                        print(prof.description())
                        swf.display(prof)
                        print(f"Candidate {x} Pareto dominates {y} but is not ranked above {y} by {swf}.")
                        print()
                    violations.append((x,y))
                    
    return violations

pareto_ranking = Axiom(
    "Pareto Ranking",
    has_violation = has_pareto_ranking_violation,
    find_all_violations = find_all_pareto_ranking_violations, 
)

def has_core_support_violation(prof,swf,verbose=False):
    """Returns True if the swf violates the "core support criterion" of https://arxiv.org/abs/2308.08430 for the given profile, False otherwise. If verbose is True, prints the profile and the core support violation.

    Args:
        prof: a Profile.
        swf (SocialWelfareFunction): An SWF to test.
        verbose (bool, default=False): If a violation is found, display the violation.

    Returns:
        Result of the test (bool): Returns True if there is a violation and False otherwise.
    """

    social_ranking = swf(prof)

    for x in prof.candidates:
        for y in prof.candidates:
            if social_ranking.extended_strict_pref(x,y):

                maj_cand_for_y = [c for c in prof.candidates if social_ranking.extended_weak_pref(c,y)]

                if isinstance(prof,Profile):
                    core_support_for_x_vs_y = [r for r in prof.rankings if r.index(x) == 0 or r.index(x) < min([r.index(c) for c in maj_cand_for_y if c != x])]     
                    core_support_for_y_vs_y = [r for r in prof.rankings if r.index(y) == 0 or r.index(y) < min([r.index(c) for c in maj_cand_for_y if c != y])]
                    core_support = core_support_for_x_vs_y + core_support_for_y_vs_y
                    restricted_prof = Profile(core_support)

                if not restricted_prof.majority_prefers(x,y):
                    if verbose:
                        print(f"Core support violation by {swf} for {x} relative to {y}:")
                        print(prof.anonymize())
                        prof.display_margin_graph()
                        print("Social ranking:",social_ranking)
                        print(f"Major candidates relative to {y}:",maj_cand_for_y)
                        print(f"Profile restricted to voters in core support for {x} relative to {y} and for {y} relative to {y}:")
                        print(restricted_prof.anonymize())
                        restricted_prof.display_margin_graph()
                    return True
    return False

def find_all_core_support_violations(prof,swf,verbose=False):
    """Returns a list of all pairs of candidates for which the swf violates the "core support criterion" of https://arxiv.org/abs/2308.08430 for the given profile. If verbose is True, prints the profile and the core support violation.

    Args:
        prof: a Profile.
        swf (SocialWelfareFunction): An SWF to test.
        verbose (bool, default=False): If a violation is found, display the violation.

    Returns:
        List of violations (list): List of all pairs of candidates for which there is a violation.
    """

    social_ranking = swf(prof)

    violations = []

    for x in prof.candidates:
        for y in prof.candidates:
            if social_ranking.extended_strict_pref(x,y):

                maj_cand_for_y = [c for c in prof.candidates if social_ranking.extended_weak_pref(c,y)]

                if isinstance(prof,Profile):
                    core_support_for_x_vs_y = [r for r in prof.rankings if r.index(x) == 0 or r.index(x) < min([r.index(c) for c in maj_cand_for_y if c != x])]     
                    core_support_for_y_vs_y = [r for r in prof.rankings if r.index(y) == 0 or r.index(y) < min([r.index(c) for c in maj_cand_for_y if c != y])]
                    core_support = core_support_for_x_vs_y + core_support_for_y_vs_y
                    restricted_prof = Profile(core_support)

                if not restricted_prof.majority_prefers(x,y):
                    if verbose:
                        print(f"Core support violation by {swf} for {x} relative to {y}:")
                        print(prof.anonymize())
                        prof.display_margin_graph()
                        print("Social ranking:",social_ranking)
                        print(f"Major candidates relative to {y}:",maj_cand_for_y)
                        print(f"Profile restricted to voters in core support for {x} relative to {y} and for {y} relative to {y}:")
                        print(restricted_prof.anonymize())
                        restricted_prof.display_margin_graph()
                    violations.append((x,y))
    return violations

core_support = Axiom(
    "Core Support",
    has_violation = has_core_support_violation,
    find_all_violations = find_all_core_support_violations, 
)



import numpy as np
from scipy.spatial import distance
from numba import jit, float32 

@jit(nopython=True, fastmath=True)
def mixed_rm_utility(v_pos: float32[:], c_pos: float32[:], beta = 0.5):
    """Based on the Rabinowitz and Macdonald (1989) mixed model described on pages 43-44 of "A Unified Theory of Voting" by S. Merrill III and B. Grofman.

    beta = 1 is the proximity quadratic utility function
    beta = 0 is the RM directional utility function

    Args:
        v_pos (numpy array): The position(s) of the voter.
        c_pos (numpy array): The position(s) of the candidate.
        beta (float): The beta parameter of the mixed model.
    
    Returns:
        float: The utility of the candidate to the voter.
    """
    return 2 * (1-beta) * np.dot(v_pos, c_pos) - beta * np.linalg.norm(v_pos - c_pos) ** 2 

def rm_utility(v_pos: float32[:], c_pos: float32[:]): 
    """Based on the Rabinowitz and Macdonald (1989) pure directional model. See "A Unified Theory of Voting" by S. Merrill III and B. Grofman, pg. 31.  

    Args:
        v_pos (numpy array): The position(s) of the voter.
        c_pos (numpy array): The position(s) of the candidate.

    Returns:
        float: The utility of the candidate to the voter.
    """

    return np.dot(v_pos, c_pos)

def linear_utility(v_pos: float32[:], c_pos: float32[:]):
    """
    The utility of the candidate for the voter is negative of the Euclidean distance between the positions. 

    Args:   
        v_pos (numpy array): The position(s) of the voter.
        c_pos (numpy array): The position(s) of the candidate.  
    Returns:
        float: The utility of the candidate to the voter.
    """
    return -np.linalg.norm(v_pos - c_pos)

def quadratic_utility(v_pos: float32[:], c_pos: float32[:]):
    """ 
    The utility of the candidate for the voter is negative of the squared Euclidean distance between the positions. 

    Args:
        v_pos (numpy array): The position(s) of the voter.
        c_pos (numpy array): The position(s) of the candidate.
    Returns:
        float: The utility of the candidate to the voter.
    """
    return -np.linalg.norm(v_pos - c_pos)**2


def city_block_utility(v_pos: float32[:], c_pos: float32[:]):
    """
    The utility of the candidate for the voter is the negative of the city-block distance between the positions (also known as the Manhattan distance). 

    Args:
        v_pos (numpy array): The position(s) of the voter.
        c_pos (numpy array): The position(s) of the candidate.
    Returns:
        float: The utility of the candidate to the voter.
    """
    return -distance.cityblock(v_pos, c_pos)

@jit(nopython=True, fastmath=True)
def shepsle_utility(v_pos: float32[:], c_pos: float32[:], kappa: float32 = 1):
    """
    The Shepsle utility function from "The Strategy of Ambiguity: Uncertainty and Electoral Competition" by Kenneth A. Shepsle, American Political Science Review, 1972, vol. 66, issue 2, pp. 555-568.   For a justification of this utility function, see Appendix B from *Making Multicandidate Elections More Democratic* (https://doi.org/10.1515/9781400859504.114) by S. Merrill III. 

    Args:
        v_pos (numpy array): The position(s) of the voter.
        c_pos (numpy array): The position(s) of the candidate. 
        kappa (float): A parameter that determines the steepness of the utility function. 
    Returns:    
        float: The utility of the candidate to the voter.      
    """
    d = np.linalg.norm(v_pos - c_pos)
    return np.exp((kappa**2 * -d**2) / 2)


@jit(nopython=True, fastmath=True)
def matthews_utility(v_pos: float32[:], c_pos: float32[:]):
    """
    Based on the Matthews directional model.  See "A Unified Theory of Voting" by S. Merrill III and B. Grofman, pg. 26.
    
    Args:
        v_pos (numpy array): The position(s) of the voter.
        c_pos (numpy array): The position(s) of the candidate.  
    Returns:    
        float: The utility of the candidate to the voter.
        
    """
    return np.dot(v_pos, c_pos) / (np.linalg.norm(v_pos) * np.linalg.norm(c_pos))




'''
    File: social_welfare_function.py
    Author: Wes Holliday (wesholliday@berkeley.edu) and Eric Pacuit (epacuit@umd.edu)
    Date: February 6, 2024
    
    The SWF class and helper functions for social welfare functions
'''

import functools

class SocialWelfareFunction(object): 
    """
    A class to add functionality to social welfare functions 

    Args:
        swf (function): An implementation of a voting method. The function should accept any type of profile, and a keyword parameter ``curr_cands`` to find the winner after restricting to ``curr_cands``. 
        name (string): The Human-readable name of the social welfare function.

    Returns:
        A ranking (Ranking) of the candidates.
    """
    def __init__(self, swf, name = None): 
        
        self.swf = swf
        self.name = name
        functools.update_wrapper(self, swf)   

    def __call__(self, edata, curr_cands = None, **kwargs):

        if (curr_cands is not None and len(curr_cands) == 0) or len(edata.candidates) == 0: 
            return []
        return self.swf(edata, curr_cands = curr_cands, **kwargs)
        
    def winners(self, edata, curr_cands = None, **kwargs):
        """Return a sorted list of the first place candidates."""

        return sorted(self.swf(edata, curr_cands = curr_cands, **kwargs).first())
    
    def display(self, edata, curr_cands = None, **kwargs):
        """Display the result of the social welfare function."""

        ranking = self.swf(edata, curr_cands = curr_cands, **kwargs)
        print(f"{self.name} ranking is {ranking}")

    def set_name(self, new_name):
        """Set the name of the social welfare function."""

        self.name = new_name

    def __str__(self): 
        return f"{self.name}"

def swf(name = None):
    """
    A decorator used when creating a social welfare function. 
    """
    def wrapper(f):
        return SocialWelfareFunction(f, name=name)
    return wrapper



'''
    File: c1_methods.py
    Author: Wes Holliday (wesholliday@berkeley.edu) and Eric Pacuit (epacuit@umd.edu)
    Date: January 10, 2022
    Update: July 31, 2022
    
    Implementations of voting methods that work on both profiles and majority graphs.
'''

from pref_voting.voting_method import  *
from pref_voting.helper import get_mg, get_weak_mg
from pref_voting.margin_based_methods import distance_to_margin_graph
from pref_voting.probabilistic_methods import c1_maximal_lottery
from pref_voting.rankings import Ranking, break_ties_alphabetically
from pref_voting.social_welfare_function import swf
import copy
import math
from itertools import product, permutations, combinations, chain
import networkx as nx
import matplotlib.pyplot as plt
from pref_voting.voting_method_properties import ElectionTypes

@vm(name = "Condorcet",
    input_types = [ElectionTypes.PROFILE, ElectionTypes.PROFILE_WITH_TIES, ElectionTypes.MAJORITY_GRAPH, ElectionTypes.MARGIN_GRAPH])
def condorcet(edata, curr_cands = None):
    """
    Return the Condorcet winner if one exists, otherwise return all the candidates.  A Condorcet winner is a candidate :math:`c` that is majority preferred to every other candidate. 

    Args:
        edata (Profile, ProfileWithTies, MajorityGraph, MarginGraph): Any election data that has a `condorcet_winner` method. 
        curr_cands (List[int], optional): If set, then find the winners for the profile restricted to the candidates in ``curr_cands``

    Returns: 
        A sorted list of candidates

    .. seealso::

        :meth:`pref_voting.profiles.Profile.condorcet_winner`,  :meth:`pref_voting.profiles_with_ties.ProfileWithTies.condorcet_winner`, :meth:`pref_voting.weighted_majority_graphs.MajorityGraph.condorcet_winner`

    :Example: 

    .. exec_code:: 

        from pref_voting.profiles import Profile
        from pref_voting.c1_methods import condorcet
        
        prof = Profile([[0, 1, 2], [1, 2, 0], [2, 0, 1]], [1, 1, 1])

        prof.display()
        print(prof.condorcet_winner())
        condorcet.display(prof)
        condorcet.display(prof.majority_graph())
        condorcet.display(prof.margin_graph())

        prof2 = Profile([[0, 1, 2], [2, 1, 0], [1, 0, 2]], [3, 1, 1])

        prof2.display()
        print(prof2.condorcet_winner())
        condorcet.display(prof2)
        condorcet.display(prof2.majority_graph())
        condorcet.display(prof2.margin_graph())

    """
   
    candidates = edata.candidates if curr_cands is None else curr_cands
    cond_winner = edata.condorcet_winner(curr_cands = curr_cands)
    
    return [cond_winner] if cond_winner is not None else sorted(candidates)

@vm(name = "Weak Condorcet",
    input_types = [ElectionTypes.PROFILE, ElectionTypes.PROFILE_WITH_TIES, ElectionTypes.MAJORITY_GRAPH, ElectionTypes.MARGIN_GRAPH])
def weak_condorcet(edata, curr_cands = None):
    
    """
    Return all weak Condorcet winner if one exists, otherwise return all the candidates.  A weak Condorcet winner is a candidate :math:`c` such that no other candidate is majority preferred to :math:`c`.

    Args:
        edata (Profile, ProfileWithTies, MajorityGraph, MarginGraph): Any election data that has a `weak_condorcet_winner` method. 
        curr_cands (List[int], optional): If set, then find the winners for the profile restricted to the candidates in ``curr_cands``

    Returns: 
        A sorted list of candidates

    .. seealso::

        :meth:`pref_voting.profiles.Profile.weak_condorcet_winner`,  
        :meth:`pref_voting.profiles_with_ties.ProfileWithTies.weak_condorcet_winner`, 
        :meth:`pref_voting.weighted_majority_graphs.MajorityGraph.weak_condorcet_winner`

    """
   
    candidates = edata.candidates if curr_cands is None else curr_cands
    weak_cond_winners = edata.weak_condorcet_winner(curr_cands = curr_cands)
    
    return weak_cond_winners if weak_cond_winners is not None else sorted(candidates)

@vm(name = "Copeland",
    input_types = [ElectionTypes.PROFILE, ElectionTypes.PROFILE_WITH_TIES, ElectionTypes.MAJORITY_GRAPH, ElectionTypes.MARGIN_GRAPH])
def copeland(edata, curr_cands = None):
    """The Copeland score for c is the number of candidates that c is majority preferred to minus the number of candidates majority preferred to c.  The Copeland winners are the candidates with the maximum Copeland score in the profile restricted to ``curr_cands``. 

    Equivalently, give each candidate 1 point for each head-to-head win, 1/2 point for each head-to-head tie, and 0 points for each head-to-head loss. Then the Copeland winners are the candidates with the maximum number of points.

    Args:
        edata (Profile, ProfileWithTies, MajorityGraph, MarginGraph): Any election data that has a `copeland_scores` method. 
        curr_cands (List[int], optional): If set, then find the winners for the profile restricted to the candidates in ``curr_cands``

    Returns: 
        A sorted list of candidates

    .. seealso::

        :meth:`pref_voting.profiles.Profile.copeland_scores`,  :meth:`pref_voting.profiles_with_ties.ProfileWithTies.copeland_scores`, :meth:`pref_voting.weighted_majority_graphs.MajorityGraph.copeland_scores`


    :Example: 
        
    .. plot:: margin_graphs_examples/mg_ex_copeland_llull.py
        :context: reset  
        :include-source: True


    .. code-block:: 

        from pref_voting.c1_methods import copeland
        copeland.display(prof)


    .. exec_code:: 
        :hide_code:

        from pref_voting.profiles import Profile
        from pref_voting.c1_methods import copeland
        
        prof = Profile([[1, 3, 0, 4, 2], [0, 1, 4, 2, 3], [2, 4, 0, 1, 3], [3, 0, 2, 4, 1],  [4, 3, 1, 0, 2], [2, 3, 0, 1, 4]], [1, 1, 1, 1, 1, 1])
        
        copeland.display(prof)
        print(prof.copeland_scores())


    """    
    c_scores = edata.copeland_scores(curr_cands = curr_cands)
    max_score = max(c_scores.values())
    return sorted([c for c in c_scores.keys() if c_scores[c] == max_score])

@swf(name = "Copeland ranking")
def copeland_ranking(edata, curr_cands=None, local=True, tie_breaking=None):
    """The SWF that ranks candidates by their Copeland scores. If local is True, then the Copeland scores are computed with respect to the profile restricted to curr_cands. Otherwise, the Copeland scores are computed with respect to the entire profile.

    Args:
        profile (Profile): An anonymous profile of linear orders on a set of candidates
        curr_cands (List[int], optional): The candidates to rank. If None, then all candidates in profile are ranked
        local (bool, optional): If True, then the Copeland scores are computed with respect to the profile restricted to curr_cands. Otherwise, the Copeland scores are computed with respect to the entire profile.
        tie_breaking (str, optional): The tie-breaking method to use. If None, then no tie-breaking is used. If "alphabetic", then the tie-breaking is done alphabetically.

    Returns:
        A Ranking object
    """

    cands = edata.candidates if curr_cands is None else curr_cands

    if local:
        copeland_scores_dict = edata.copeland_scores(curr_cands=cands)

    else:
        c_scores = edata.copeland_scores(curr_cands=edata.candidates)
        copeland_scores_dict = {c: c_scores[c] for c in cands}

    for cand in cands:
        copeland_scores_dict[cand] = -copeland_scores_dict[cand]

    copeland_ranking = Ranking(copeland_scores_dict)
    copeland_ranking.normalize_ranks()

    if tie_breaking == "alphabetic":
        copeland_ranking = break_ties_alphabetically(copeland_ranking)

    return copeland_ranking


@vm(name = "Llull",
    input_types = [ElectionTypes.PROFILE, ElectionTypes.PROFILE_WITH_TIES, ElectionTypes.MAJORITY_GRAPH, ElectionTypes.MARGIN_GRAPH])
def llull(edata, curr_cands = None):
    """The Llull score for a candidate :math:`c` is the number of candidates that :math:`c` is weakly majority preferred to.  This is equivalent to calculating the Copeland scores for a candidate :math:`c` with 1 point for each candidate that :math:`c` is majority preferred to, 1 point for each candidate that :math:`c` is tied with (instead of 1/2 a point as for Copeland), and 0 points for each candidate that is majority preferred to :math:`c`.  The Llull winners are the candidates with the maximum Llull score in the profile restricted to ``curr_cands``. 

    Args:
        edata (Profile, ProfileWithTies, MajorityGraph, MarginGraph): Any election data that has a `copeland_scores` method. 
        curr_cands (List[int], optional): If set, then find the winners for the profile restricted to the candidates in ``curr_cands``

    Returns: 
        A sorted list of candidates

    .. seealso::

        :meth:`pref_voting.profiles.Profile.copeland_scores`,  :meth:`pref_voting.profiles_with_ties.ProfileWithTies.copeland_scores`, :meth:`pref_voting.weighted_majority_graphs.MajorityGraph.copeland_scores`

    :Example: 
        
    .. plot::  margin_graphs_examples/mg_ex_copeland_llull.py
        :context: reset  
        :include-source: True


    .. code-block:: 

        from pref_voting.c1_methods import llull
        llull.display(prof)


    .. exec_code:: 
        :hide_code:

        from pref_voting.profiles import Profile
        from pref_voting.c1_methods import llull
        
        prof = Profile([[1, 3, 0, 4, 2], [0, 1, 4, 2, 3], [2, 4, 0, 1, 3], [3, 0, 2, 4, 1],  [4, 3, 1, 0, 2], [2, 3, 0, 1, 4]], [1, 1, 1, 1, 1, 1])
        
        llull.display(prof)
        print(prof.copeland_scores(scores=(1, 0.5, 0)))

    """ 

    l_scores = edata.copeland_scores(curr_cands = curr_cands, scores = (1,1,0))
    max_score = max(l_scores.values())
    return sorted([c for c in l_scores.keys() if l_scores[c] == max_score])

def left_covers(dom, c1, c2):
    # left covers: c1 left covers c2 when all the candidates that are majority preferred to c1 are also majority preferred to c2.
    
    # weakly left covers: c1 weakly left covers c2 when all the candidates that are majority preferred to or tied with c1
    # are also majority preferred to or tied with c2.
    
    return dom[c1].issubset(dom[c2])

def right_covers(dom, c1, c2):
    # right covers: c1 right covers c2 when all the candidates that c2  majority preferrs are majority
    # preferred by c1
      
    return dom[c2].issubset(dom[c1])

@vm(name = "Uncovered Set",
    input_types = [ElectionTypes.PROFILE, ElectionTypes.PROFILE_WITH_TIES, ElectionTypes.MAJORITY_GRAPH, ElectionTypes.MARGIN_GRAPH])
def uc_gill(edata, curr_cands = None): 
    """Uncovered Set (Gillies version):  Given candidates :math:`a` and :math:`b`, say that :math:`a` defeats :math:`b` in the election if :math:`a` is majority preferred to :math:`b` and :math:`a` left covers :math:`b`: i.e., for all :math:`c`, if :math:`c` is majority preferred to :math:`a`,  then :math:`c` majority preferred to :math:`b`. The winners are the set of candidates who are undefeated in the election restricted to ``curr_cands``. 
    
    Args:
        edata (Profile, ProfileWithTies, MajorityGraph, MarginGraph): Any election data that has a `dominators` method. 
        curr_cands (List[int], optional): If set, then find the winners for the profile restricted to the candidates in ``curr_cands``

    Returns: 
        A sorted list of candidates

    .. seealso::

        :func:`pref_voting.c1_methods.uc_fish`,  :func:`pref_voting.c1_methods.uc_bordes`, :func:`pref_voting.c1_methods.uc_mckelvey`

    :Example: 
        
    .. plot::  margin_graphs_examples/mg_ex_uncovered_sets.py
        :context: reset  
        :include-source: True

    .. code-block:: 

        from pref_voting.c1_methods import uc_gill
        uc_gill.display(prof)


    .. exec_code:: 
        :hide_code:

        from pref_voting.profiles import Profile
        from pref_voting.c1_methods import uc_gill
        
        prof = Profile([[2, 3, 0, 1], [0, 2, 1, 3], [3, 0, 1, 2], [1, 2, 0, 3], [1, 2, 3, 0]], [1, 1, 1, 2, 1])
        
        uc_gill.display(prof)
    
    """

    candidates = edata.candidates if curr_cands is None else curr_cands
    dom = {c: set(edata.dominators(c, curr_cands = curr_cands)) for c in candidates}
    uc_set = list()
    for c1 in candidates:
        is_in_ucs = True
        for c2 in edata.dominators(c1, curr_cands = curr_cands): # consider only c2 predecessors
            if c1 != c2:
                # check if c2 left covers  c1 
                if left_covers(dom, c2, c1):
                    is_in_ucs = False
        if is_in_ucs:
            uc_set.append(c1)
    return list(sorted(uc_set))

def uc_gill_defeat(edata, curr_cands = None): 
    """Returns the defeat relation used to find the  Uncovered Set (Gillies version) winners.
    
    Args:
        edata (Profile, ProfileWithTies, MajorityGraph, MarginGraph): Any election data that has a `dominators` method. 
        curr_cands (List[int], optional): If set, then find the winners for the profile restricted to the candidates in ``curr_cands``

    Returns: 
        A networkx object in which there is an edge from :math:`a` to :math:`b` when :math:`a` to :math:`b` according to Top Cycle. 

    .. seealso::

        :func:`pref_voting.c1_methods.uc_gill`

    :Example: 
        
        
    .. plot::  margin_graphs_examples/uc_gill_defeat_example.py
        :include-source: True

    
    """
    
    defeat = nx.DiGraph()

    candidates = edata.candidates if curr_cands is None else curr_cands
    
    defeat.add_nodes_from(candidates)
    
    dom = {c: set(edata.dominators(c, curr_cands = curr_cands)) for c in candidates}
    for c1 in candidates:
        for c2 in edata.dominators(c1, curr_cands = curr_cands): # consider only c2 predecessors
            if c1 != c2:
                # check if c2 left covers  c1 
                if left_covers(dom, c2, c1):
                    defeat.add_edge(c2, c1)
    return defeat

@vm(name = "Uncovered Set - Fishburn",
    input_types = [ElectionTypes.PROFILE, ElectionTypes.PROFILE_WITH_TIES, ElectionTypes.MAJORITY_GRAPH, ElectionTypes.MARGIN_GRAPH])
def uc_fish(edata, curr_cands = None): 
    """Uncovered Set (Fishburn version):  Given candidates :math:`a` and :math:`b`, say that :math:`a` defeats :math:`b` in the election :math:`a` left covers :math:`b`: i.e., for all :math:`c`, if :math:`c` is majority preferred to :math:`a`,  then :math:`c` majority preferred to :math:`b`. The winners are the set of candidates who are undefeated in the election restricted to ``curr_cands``. 
    
    Args:
        edata (Profile, ProfileWithTies, MajorityGraph, MarginGraph): Any election data that has a `dominators` method. 
        curr_cands (List[int], optional): If set, then find the winners for the profile restricted to the candidates in ``curr_cands``

    Returns: 
        A sorted list of candidates

    .. seealso::

        :func:`pref_voting.c1_methods.uc_gill`,  :func:`pref_voting.c1_methods.uc_bordes`, :func:`pref_voting.c1_methods.uc_mckelvey`

    :Example: 
        
        
    .. plot::  margin_graphs_examples/mg_ex_uncovered_sets.py
        :include-source: True


    .. code-block:: 

        from pref_voting.c1_methods import uc_fish
        uc_fish.display(prof)


    .. exec_code:: 
        :hide_code:

        from pref_voting.profiles import Profile
        from pref_voting.c1_methods import uc_fish
        
        prof = Profile([[2, 3, 0, 1], [0, 2, 1, 3], [3, 0, 1, 2], [1, 2, 0, 3], [1, 2, 3, 0]], [1, 1, 1, 2, 1])
        
        uc_fish.display(prof)
    
    """
    candidates = edata.candidates if curr_cands is None else curr_cands
    dom = {c: set(edata.dominators(c, curr_cands = curr_cands)) for c in candidates}
    uc_set = list()
    for c1 in candidates:
        is_in_ucs = True
        for c2 in candidates:
            if c1 != c2:
                # check if c2 left covers  c1 but c1 does not left cover c2
                if left_covers(dom, c2, c1)  and not left_covers(dom, c1, c2):
                    is_in_ucs = False
        if is_in_ucs:
            uc_set.append(c1)
    return list(sorted(uc_set))

def uc_fish_defeat(edata, curr_cands = None): 
    """Returns the defeat relation used to find the  Uncovered Set (Fishburn version) winners.
    
    Args:
        edata (Profile, ProfileWithTies, MajorityGraph, MarginGraph): Any election data that has a `dominators` method. 
        curr_cands (List[int], optional): If set, then find the winners for the profile restricted to the candidates in ``curr_cands``

    Returns: 
        A networkx object in which there is an edge from :math:`a` to :math:`b` when :math:`a` to :math:`b` according to Top Cycle. 

    .. seealso::

        :func:`pref_voting.c1_methods.uc_fish`


    :Example: 
        
        
    .. plot::  margin_graphs_examples/uc_fish_defeat_example.py
        :include-source: True

    """
    defeat = nx.DiGraph()

    candidates = edata.candidates if curr_cands is None else curr_cands
    
    defeat.add_nodes_from(candidates)
    dom = {c: set(edata.dominators(c, curr_cands = curr_cands)) for c in candidates}
    for c1 in candidates:
        is_in_ucs = True
        for c2 in candidates:
            if c1 != c2:
                # check if c2 left covers  c1 but c1 does not left cover c2
                if left_covers(dom, c2, c1)  and not left_covers(dom, c1, c2):
                    defeat.add_edge(c2, c1)
    return defeat

@vm(name = "Uncovered Set - Bordes",
    input_types = [ElectionTypes.PROFILE, ElectionTypes.PROFILE_WITH_TIES, ElectionTypes.MAJORITY_GRAPH, ElectionTypes.MARGIN_GRAPH])
def uc_bordes(edata, curr_cands = None): 
    """Uncovered Set (Bordes version):  Given candidates :math:`a` and :math:`b`, say that :math:`a` Bordes covers :math:`b` if :math:`a` is majority preferred to :math:`b` and for all :math:`c`, if :math:`c` is majority preferred or tied with :math:`a`, then :math:`c` is majority preferred to or tied with :math:`b`. The winners are the set of candidates who are not Bordes covered in the election restricted to ``curr_cands``. 
    
    Args:
        edata (Profile, ProfileWithTies, MajorityGraph, MarginGraph): Any election data that has  `dominators` and `majority_prefers` methods. 
        curr_cands (List[int], optional): If set, then find the winners for the profile restricted to the candidates in ``curr_cands``

    Returns: 
        A sorted list of candidates

    .. seealso::

        :func:`pref_voting.c1_methods.uc_gill`,  :func:`pref_voting.c1_methods.uc_fish`, :func:`pref_voting.c1_methods.uc_mckelvey`

    :Example: 
        
        
    .. plot::  margin_graphs_examples/mg_ex_uncovered_sets.py
        :context: reset  
        :include-source: True


    .. code-block:: 

        from pref_voting.c1_methods import uc_bordes
        uc_bordes.display(prof)


    .. exec_code:: 
        :hide_code:

        from pref_voting.profiles import Profile
        from pref_voting.c1_methods import uc_bordes
        
        prof = Profile([[2, 3, 0, 1], [0, 2, 1, 3], [3, 0, 1, 2], [1, 2, 0, 3], [1, 2, 3, 0]], [1, 1, 1, 2, 1])
        
        uc_bordes.display(prof)
    
    """

    candidates = edata.candidates if curr_cands is None else curr_cands

    dom = {c: set(edata.dominators(c, curr_cands = curr_cands)).union([_c for _c in candidates if (not edata.majority_prefers(c, _c) and not edata.majority_prefers(_c, c))]) for c in candidates}
    
    uc_set = list()
    for c1 in candidates:
        is_in_ucs = True
        for c2 in edata.dominators(c1, curr_cands = curr_cands): # consider only c2 predecessors
            if c1 != c2:
                # check if c2 left covers  c1 
                if left_covers(dom, c2, c1):
                    is_in_ucs = False
        if is_in_ucs:
            uc_set.append(c1)
    return list(sorted(uc_set))  

@vm(name = "Uncovered Set - McKelvey",
    input_types = [ElectionTypes.PROFILE, ElectionTypes.PROFILE_WITH_TIES, ElectionTypes.MAJORITY_GRAPH, ElectionTypes.MARGIN_GRAPH])
def uc_mckelvey(edata, curr_cands = None): 
    """Uncovered Set (McKelvey version):  Given candidates :math:`a` and :math:`b`, say that  :math:`a` McKelvey covers :math:`b` if a Gillies covers :math:`b` and :math:`a` Bordes covers :math:`b`. The winners are the set of candidates who are not McKelvey covered in the election restricted to ``curr_cands``. 
    
    Args:
        edata (Profile, ProfileWithTies, MajorityGraph, MarginGraph): Any election data that has  `dominators` and `majority_prefers` methods. 
        curr_cands (List[int], optional): If set, then find the winners for the profile restricted to the candidates in ``curr_cands``

    Returns: 
        A sorted list of candidates

    .. seealso::

        :func:`pref_voting.c1_methods.uc_gill`, :func:`pref_voting.c1_methods.uc_fish`, :func:`pref_voting.c1_methods.uc_bordes`

    :Example: 
         
    .. plot::  margin_graphs_examples/mg_ex_uncovered_sets.py
        :context: reset  
        :include-source: True


    .. code-block:: 

        from pref_voting.c1_methods import uc_mckelvey
        uc_bordes.display(prof)


    .. exec_code:: 
        :hide_code:

        from pref_voting.profiles import Profile
        from pref_voting.c1_methods import uc_mckelvey
        
        prof = Profile([[2, 3, 0, 1], [0, 2, 1, 3], [3, 0, 1, 2], [1, 2, 0, 3], [1, 2, 3, 0]], [1, 1, 1, 2, 1])
        
        uc_mckelvey.display(prof)
    
    """
    candidates = edata.candidates if curr_cands is None else curr_cands

    strict_dom = {c: set(edata.dominators(c, curr_cands = curr_cands)) for c in candidates}    
    weak_dom = {c: strict_dom[c].union([_c for _c in candidates if (not edata.majority_prefers(c, _c) and not edata.majority_prefers(_c, c))]) for c in candidates}
    uc_set = list()
    for c1 in candidates:
        is_in_ucs = True
        for c2 in edata.dominators(c1, curr_cands = curr_cands): # consider only c2 predecessors
            if c1 != c2:
                # check if c2 left covers  c1 
                if left_covers(strict_dom, c2, c1) and left_covers(weak_dom, c2, c1):
                    is_in_ucs = False
        if is_in_ucs:
            uc_set.append(c1)
    return list(sorted(uc_set))      

@vm(name = "Top Cycle",
    input_types = [ElectionTypes.PROFILE, ElectionTypes.PROFILE_WITH_TIES, ElectionTypes.MAJORITY_GRAPH, ElectionTypes.MARGIN_GRAPH])
def top_cycle(edata, curr_cands = None):
    """The smallest set of candidates such that every candidate inside the set is majority preferred to every candidate outside the set.  
    
    Args:
        edata (Profile, ProfileWithTies, MajorityGraph, MarginGraph): Any election data that has a `majority_prefers` method. 
        curr_cands (List[int], optional): If set, then find the winners for the profile restricted to the candidates in ``curr_cands``

    Returns: 
        A sorted list of candidates

    .. seealso::

        Also known as ``getcha`` and ``smith_set``. 
        
        Related function includes :func:`pref_voting.c1_methods.gocha`

    :Example: 
        
        
    .. plot::  margin_graphs_examples/mg_ex_top_cycle_gocha.py
        :context: reset  
        :include-source: True


    .. code-block:: 

        from pref_voting.c1_methods import top_cycle, getcha, smith_set
        top_cycle.display(prof)
        getcha.display(prof)
        smith_set.display(prof)


    .. exec_code:: 
        :hide_code:

        from pref_voting.profiles import Profile
        from pref_voting.c1_methods import top_cycle, getcha, smith_set
        
        prof = Profile([[1, 2, 0, 3], [1, 3, 0, 2], [3, 1, 0, 2], [0, 3, 1, 2]], [1, 1, 1, 1])
        
        top_cycle.display(prof)
        getcha.display(prof)
        smith_set.display(prof)

        
    """    
    wmg = get_weak_mg(edata, curr_cands = curr_cands)
    scc = list(nx.strongly_connected_components(wmg))
    min_indegree = min([max([wmg.in_degree(n) for n in comp]) for comp in scc])
    smith = [comp for comp in scc if max([wmg.in_degree(n) for n in comp]) == min_indegree][0]
    return sorted(list(smith))

# Create some aliases for Top Cycle
top_cycle.set_name("GETCHA")
getcha = copy.deepcopy(top_cycle)
getcha.skip_registration = True

top_cycle.set_name("Smith Set")
smith_set = copy.deepcopy(top_cycle)
smith_set.skip_registration = True

# reset the name Top Cycle
top_cycle.set_name("Top Cycle")

def top_cycle_defeat(edata, curr_cands = None):
    """Return the defeat relation associated with the Top Cycle voting method. 
    
    Args:
        edata (Profile, ProfileWithTies, MajorityGraph, MarginGraph): Any election data that has a `majority_prefers` method. 
        curr_cands (List[int], optional): If set, then find the winners for the profile restricted to the candidates in ``curr_cands``

    Returns: 
        A networkx object in which there is an edge from :math:`a` to :math:`b` when :math:`a` to :math:`b` according to Top Cycle. 

    .. seealso::

        :func:`pref_voting.c1_methods.top_cycle`

    :Example: 
        
    .. plot::  margin_graphs_examples/top_cycle_defeat.py
        :context: reset  
        :include-source: True
 
    """    

    defeat = nx.DiGraph()
    candidates = edata.candidates if curr_cands is None else curr_cands
    smith_set = top_cycle(edata, curr_cands = candidates)
    
    defeat.add_nodes_from(candidates)
    defeat.add_edges_from([(a, b) for a in candidates for b in candidates if a != b and a in smith_set and b not in smith_set])
    return defeat

@vm(name = "GOCHA",
    input_types = [ElectionTypes.PROFILE, ElectionTypes.PROFILE_WITH_TIES, ElectionTypes.MAJORITY_GRAPH, ElectionTypes.MARGIN_GRAPH])
def gocha(edata, curr_cands = None):
    """The GOCHA set (also known as the Schwartz set) is the set of all candidates x such that if y can reach x in the transitive closer of the majority relation, then x can reach y in the transitive closer of the majority relation.
      
    Args:
        edata (Profile, ProfileWithTies, MajorityGraph, MarginGraph): Any election data that has a `majority_prefers` method. 
        curr_cands (List[int], optional): If set, then find the winners for the profile restricted to the candidates in ``curr_cands``

    Returns: 
        A sorted list of candidates

    .. seealso::

        Also known as ``schwartz_set``. 
        
        Related function includes :func:`pref_voting.c1_methods.top_cycle`

    :Example: 
          
    .. plot::  margin_graphs_examples/mg_ex_top_cycle_gocha.py
        :context: reset  
        :include-source: True

    .. code-block:: 

        from pref_voting.c1_methods import top_cycle, gocha, schwartz_set

        gocha.display(prof)
        schwartz_set.display(prof)

    .. exec_code:: 
        :hide_code:

        from pref_voting.profiles import Profile
        from pref_voting.c1_methods import gocha, schwartz_set
        
        prof = Profile([[1, 2, 0, 3], [1, 3, 0, 2], [3, 1, 0, 2], [0, 3, 1, 2]], [1, 1, 1, 1])
        
        gocha.display(prof)
        schwartz_set.display(prof)
    
    """    
    
    mg = get_mg(edata, curr_cands = curr_cands)
    transitive_closure =  nx.algorithms.dag.transitive_closure(mg)
    schwartz = set()
    for ssc in nx.strongly_connected_components(transitive_closure):
        if not any([transitive_closure.has_edge(c2,c1) 
                    for c1 in ssc for c2 in transitive_closure.nodes if c2 not in ssc]):
            schwartz =  schwartz.union(ssc)
    return sorted(list(schwartz))

# Create some aliases for GOCHA
gocha.set_name("Schwartz Set")
schwartz_set = copy.deepcopy(gocha)
schwartz_set.skip_registration = True

# reset the name GETCHA
gocha.set_name("GOCHA")


## Banks
#

def seqs(iterable):
    s = list(iterable)
    return chain.from_iterable(permutations(s, r) for r in range(len(s)+1))

def is_transitive(G, p):
    for c1_idx, c1 in enumerate(p[:-1]):
        for c2 in p[c1_idx+1::]:            
            if not G.has_edge(c1,c2):
                return False
    return True

def is_subsequence(x, y):
    it = iter(y)
    return all(any(c == ch for c in it) for ch in x)

@vm(name = "Banks",
    input_types = [ElectionTypes.PROFILE, ElectionTypes.PROFILE_WITH_TIES, ElectionTypes.MAJORITY_GRAPH, ElectionTypes.MARGIN_GRAPH])
def banks(edata, curr_cands = None): 
    """ Say that a *chain* in majority graph is a subset of candidates that is linearly ordered by the majority relation. Then a candidate :math:`a` if :math:`a` is the maximum element with respect to the majority relation of some maximal chain in the majority graph.

    Args:
        edata (Profile, ProfileWithTies, MarginGraph): Any election data that has a `margin` method. 
        curr_cands (List[int], optional): If set, then find the winners for the profile restricted to the candidates in ``curr_cands``

    Returns: 
        A sorted list of candidates


    :Example: 

    .. plot::  margin_graphs_examples/mg_ex_banks.py
        :context: reset  
        :include-source: True


    .. code-block:: 

        from pref_voting.c1_methods import banks

        banks.display(prof)


    .. exec_code:: 
        :hide_code:

        from pref_voting.weighted_majority_graphs import MarginGraph
        from pref_voting.c1_methods import banks
        
        mg = MarginGraph([0, 1, 2, 3], [(0, 2, 2), (0, 3, 6), (1, 0, 8), (2, 3, 4), (2, 1, 10), (3, 1, 12)])

        banks.display(mg)

    """

    mg = get_mg(edata, curr_cands = curr_cands)
    trans_paths = list()
    for s in seqs(mg.nodes):
        if nx.algorithms.simple_paths.is_simple_path(mg, s):
            if is_transitive(mg, s): 
                trans_paths.append(s)

    maximal_paths = list()
    #print("max paths")
    for s in trans_paths:
        is_max = True
        for other_s in trans_paths: 
            if s != other_s:
                if is_subsequence(s, other_s): 
                    is_max = False
                    break
        if is_max:
            maximal_paths.append(s)
    
    return sorted(list(set([p[0] for p in maximal_paths])))

def banks_with_explanation(edata, curr_cands = None): 
    """Return the Banks winners and the list of maximal chains in the majority graph. 

    Args:
        edata (Profile, ProfileWithTies, MarginGraph): Any election data that has a `margin` method. 
        curr_cands (List[int], optional): If set, then find the winners for the profile restricted to the candidates in ``curr_cands``

    Returns: 
        A sorted list of candidates

        A list of lists of candidates each representing a maximal chain in the majority graph

    :Example: 

    .. plot::  margin_graphs_examples/mg_ex_banks.py
        :context: reset  
        :include-source: True


    .. code-block:: 

        from pref_voting.c1_methods import banks_with_explanation

        bws, maximal_chains = banks_with_explanation(mg)

        print(f"Winning set: {bws}")
        for c in maximal_chains: 
            print(f"Maximal chain: {c}")


    .. exec_code:: 
        :hide_code:

        from pref_voting.weighted_majority_graphs import MarginGraph
        from pref_voting.c1_methods import banks_with_explanation
        
        mg = MarginGraph([0, 1, 2, 3], [(0, 2, 2), (0, 3, 6), (1, 0, 8), (2, 3, 4), (2, 1, 10), (3, 1, 12)])

        bws, maximal_chains = banks_with_explanation(mg)

        print(f"Winning set: {bws}")
        for c in maximal_chains: 
            print(f"Maximal chain: {c}")

    """

    mg = get_mg(edata, curr_cands = curr_cands)
    trans_paths = list()
    for s in seqs(mg.nodes):
        if nx.algorithms.simple_paths.is_simple_path(mg, s):
            if is_transitive(mg, s): 
                trans_paths.append(s)

    maximal_paths = list()
    #print("max paths")
    for s in trans_paths:
        is_max = True
        for other_s in trans_paths: 
            if s != other_s:
                if is_subsequence(s, other_s): 
                    is_max = False
                    break
        if is_max:
            maximal_paths.append(s)
    
    return sorted(list(set([p[0] for p in maximal_paths]))), maximal_paths


def lin_order_to_rel(lin_order): 
    """Convert a linear order (a list of items) into a set of ordered pairs"""
    els = sorted(lin_order)
    rel = []
    for a,b in combinations(els, 2):
        if lin_order.index(a) < lin_order.index(b): 
            rel.append((a,b))
        elif lin_order.index(b) < lin_order.index(a): 
            rel.append((b,a))     
    return rel


def slater_rankings(edata, curr_cands = None): 
    """
    A Slater ranking is a linear order :math:`R` of the candidates that minimizes the number of edges in the majority graph we have to turn around before we obtain :math:`R`. 

    Args:
        edata (Profile, ProfileWithTies, MarginGraph): Any election data that has a `margin` method. 
        curr_cands (List[int], optional): If set, then find the winners for the profile restricted to the candidates in ``curr_cands``

    Returns: 
        rankings: A list of Slater rankings.

        dist: The minimum distance of the Slater rankings.

    :Example:

    .. exec_code::

        from pref_voting.weighted_majority_graphs import MarginGraph
        from pref_voting.c1_methods import slater_rankings
        
        mg = MarginGraph([0, 1, 2, 3], [(0, 2, 2), (0, 3, 6), (1, 0, 8), (2, 3, 4), (2, 1, 10), (3, 1, 12)])

        srs, d = slater_rankings(mg)
        print(f"minimum distance: {d}")
        for sr in srs: 
            print(f"ranking: {sr}") 
    """
    candidates = edata.candidates if curr_cands is None else curr_cands
    min_dist = np.inf
    
    rankings = list()
    for lin_order in permutations(candidates): 
        lo_rel = lin_order_to_rel(lin_order)
        
        dist = distance_to_margin_graph(edata, lo_rel, exp = 0, curr_cands = curr_cands)
        if dist < min_dist: 
            min_dist = dist
            rankings = [lin_order]
        elif dist == min_dist: 
            rankings.append(lin_order)
    return rankings, min_dist

@vm(name = "Slater",
    input_types = [ElectionTypes.PROFILE, ElectionTypes.PROFILE_WITH_TIES, ElectionTypes.MAJORITY_GRAPH, ElectionTypes.MARGIN_GRAPH])
def slater(edata, curr_cands = None): 
    """A Slater ranking is a linear order :math:`R` of the candidates that minimizes the number of edges in the majority graph we have to turn around before we obtain :math:`R`.   A candidate is a Slater winner if the candidate is the top element of some Slater ranking.

    Args:
        edata (Profile, ProfileWithTies, MarginGraph): Any election data that has a `margin` method. 
        curr_cands (List[int], optional): If set, then find the winners for the profile restricted to the candidates in ``curr_cands``

    Returns: 
        A sorted list of candidates


    :Example: 

    .. plot::  margin_graphs_examples/mg_ex_slater.py
        :context: reset  
        :include-source: True


    .. code-block:: 

        from pref_voting.c1_methods import slater

        slater.display(prof)


    .. exec_code:: 
        :hide_code:

        from pref_voting.weighted_majority_graphs import MarginGraph
        from pref_voting.c1_methods import slater
        
        mg = MarginGraph([0, 1, 2, 3], [(0, 2, 2), (0, 3, 6), (1, 0, 8), (2, 3, 4), (2, 1, 10), (3, 1, 12)])

        slater.display(mg)

    """    
    rankings, dist = slater_rankings(edata, curr_cands = curr_cands)
    
    return sorted(list(set([r[0] for r in rankings])))

@vm(name = "Bipartisan Set",
    input_types = [ElectionTypes.PROFILE, ElectionTypes.PROFILE_WITH_TIES, ElectionTypes.MAJORITY_GRAPH, ElectionTypes.MARGIN_GRAPH])
def bipartisan(edata, curr_cands = None, threshold = 0.0000001): 
    """The Bipartisan Set is the support of the (chosen) C1 maximal lottery.

    Args:
        edata (Profile, ProfileWithTies, MajorityGraph, MarginGraph): Any election data that has a `margin_matrix` attribute.
        curr_cands (List[int], optional): If set, then find the winners for the profile restricted to the candidates in ``curr_cands``

    Returns:
        A sorted list of candidates.
    """

    ml = c1_maximal_lottery(edata, curr_cands=curr_cands)
    return sorted([c for c in ml.keys() if  ml[c] > threshold])

c1_swf = [
    copeland_ranking
]

defeat_methods = [
    top_cycle_defeat,
    uc_gill_defeat,
    uc_fish_defeat
]




'''
    File: grade_profiles.py
    Author: Wes Holliday (wesholliday@berkeley.edu) and Eric Pacuit (epacuit@umd.edu)
    Date: September 24, 2023
    
    Functions to reason about profiles of grades.
'''

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
from tabulate import tabulate
from pref_voting.mappings import Grade, _Mapping
from pref_voting.profiles_with_ties import ProfileWithTies

class GradeProfile(object):
    """An anonymous profile of (truncated) grades.  

    :param grade_maps: List of grades in the profile, where a grade is either a :class:`Grade` object or a dictionary.
    :type grade_maps: list[dict[int or str: int or str]] or list[Grade]
    :param grades: List of grades.
    :type gcounts: list[int or str]
    :param gcounts: List of the number of voters associated with each grade.  Should be the same length as grade_maps.  If not provided, it is assumed that 1 voter submitted each element of ``grade_maps``.
    :type gcounts: list[int], optional
    :param candidates: List of candidates in the profile.  If not provided, it is the candidates that are assigned a grade by least on voter.
    :type candidates: list[int] or list[str], optional
    :param cmap: Dictionary mapping candidates to candidate names (strings).  If not provided, each candidate name is mapped to itself.
    :type cmap: dict[int or str: str], optional
    :param gmap: Dictionary mapping grades to grade names (strings).  If not provided, each grade is mapped to itself.
    :type gmap: dict[int or str: str], optional
    :param grade_order: A list of the grades representing the order of the grades. It is assumed the grades are listed from largest to smallest.  If not provided, the grades are assumed to be numbers and compared using the greater-than relation.
    :type gmap: list[int or str], optional

    :Example:

    The following code creates a profile in which
    2 voters submitted the ranking 0 ranked first, 1 ranked second, and 2 ranked third; 3 voters submitted the ranking 1 and 2 are tied for first place and 0 is ranked second; and 1 voter submitted the ranking in which 2 is ranked first and 0 is ranked second:

    .. code-block:: python

        gprof =  GradeProfile([{"x":1, "y":3, "z":1}, {"x":0, "y":-1, "z":3}, {"x":0, "y":-1}, {"x":0, "y":1, "z":2}], [-1, 0, 1, 2, 3], gcounts=[2, 3, 1, 1], candidates=["x", "y", "z"])

        gprof.display()

    """

    def __init__(
        self, 
        grade_maps, 
        grades, 
        gcounts=None, 
        candidates=None, 
        cmap=None,
        gmap=None,
        grade_order=None):
        """Constructor method"""

        assert gcounts is None or len(grade_maps) == len(
            gcounts
        ), "The number of grades much be the same as the number of gcounts"
        self.candidates = (
            sorted(candidates)
            if candidates is not None
            else sorted(list(set([x for g in grade_maps for x in g.keys()])))
        )
        """The domain of the profile. """
        self.cmap = cmap if cmap is not None else {c: str(c) for c in self.candidates}
        """The candidate map is a dictionary associating an alternative with the name used when displaying a alternative."""

        self.grades=grades
        """The grades in the profile. """

        self.can_sum_grades = all([isinstance(g, (float, int)) for g in self.grades])

        self.grade_order = grade_order if grade_order is not None else sorted(self.grades, reverse = True)
        """The order of the grades. If None, then order from largest to smallest"""

        self.use_grade_order = grade_order is not None

        self.compare_function = lambda v1, v2: (v1 > v2) - (v2 > v1) if grade_order is None else lambda v1, v2: (grade_order.index(v1) < grade_order.index(v2)) - (grade_order.index(v2) < grade_order.index(v1))

        self.gmap = gmap if gmap is not None else {g: str(g) for g in self.grades}
        """The candidate map is a dictionary associating an alternative with the name used when displaying a alternative."""

        self._grades = [
            Grade(g_map, self.grades, candidates=self.candidates, cmap=self.cmap, gmap=self.gmap, compare_function=self.compare_function)
            if type(g_map) == dict
            else Grade(g_map.as_dict(), self.grades, candidates=self.candidates, cmap=self.cmap, gmap=self.gmap, compare_function=self.compare_function)
            for g_map in grade_maps
        ]
        """The list of grades in the Profile (each utility is a :class:`Utility` object). 
        """
        self.gcounts = [1] * len(grade_maps) if gcounts is None else list(gcounts)

        self.num_voters = np.sum(self.gcounts)
        """The number of voters in the profile. """

    @property
    def grades_counts(self):
        """Returns the grade and the counts of each grade."""

        return self._grades, self.gcounts
    
    @property
    def grade_functions(self):
        """Return all of the grade functions in the profile."""
        
        gs = list()
        for g,c in zip(self._grades, self.gcounts): 
            gs += [g] * c
        return gs

    def has_grade(self, c):
        """Return True if ``c`` is assigned a grade by at least one voter."""

        return any([g.has_grade(c) for g in self._grades])

    @property
    def num_cands(self):
        """Return the number of candidates in the profile."""
        return len(self.candidates)
    
    def margin(self, c1, c2, use_extended=False): 
        """
        Return the margin of ``c1`` over ``c2``.  If ``c1`` is not assigned a grade by any voter, return None.
        """
        if use_extended: 
            return np.sum([num for g,num in zip(*self.grades_counts) if g.extended_strict_pref(c1, c2)]) - np.sum([num for g,num in zip(*self.grades_counts) if g.extended_strict_pref(c2, c1)])
        else: 
            return np.sum([num for g,num in zip(*self.grades_counts) if g.strict_pref(c1, c2)]) - np.sum([num for g,num in zip(*self.grades_counts) if g.strict_pref(c2, c1)])

    def proportion(self, cand, grade):
        """
        Return the proportion of voters that assign ``cand`` the grade ``grade``.

        Note that ``grade`` could be None, in which case the proportion of voters that do not assign ``cand`` a grade is returned.
        """
        return np.sum([num for g,num in zip(*self.grades_counts) if g(cand) == grade]) / self.num_voters

    def sum(self, c): 
        """Return the sum of the grades of ``c``.  If ``c`` is not assigned a grade by any voter, return None."""

        assert self.can_sum_grades, "The grades in the profile cannot be summed."

        return np.sum([g(c) * num for g,num in zip(*self.grades_counts) if g.has_grade(c)]) if self.has_grade(c) else None
    
    def avg(self, c): 
        """Return the average of the grades of ``c``.  If ``c`` is not assigned a grade by any voter, return None."""

        assert self.can_sum_grades, "The grades in the profile cannot be summed."

        return np.mean([g(c) for g in self.grade_functions if g.has_grade(c)]) if self.has_grade(c) else None
    
    def max(self, c): 
        """Return the maximum of the grade of ``c``.  If ``c`` is not assigned a grade by any voter, return None."""

        grades_for_c = [-1 * self.grade_order.index(g(c)) for g in self._grades if g.has_grade(c)] if self.use_grade_order else [g(c) for g in self._grades if g.has_grade(c)]

        return (self.grade_order[-1 * max(grades_for_c)] if self.use_grade_order else max(grades_for_c)) if self.has_grade(c) else None
    
    def min(self, c): 
        """Return the minimum of the grades of ``c``.  If ``c`` is not assigned a grade by any voter, return None."""

        grades_for_c = [-1 * self.grade_order.index(g(c)) for g in self._grades if g.has_grade(c)] if self.use_grade_order else [g(c) for g in self._grades if g.has_grade(c)]

        return (self.grade_order[-1 * min(grades_for_c)] if self.use_grade_order else min(grades_for_c)) if self.has_grade(c) else None
    
    def median(self, c, use_lower=True, use_average=False): 
        """Return the median of the grades of ``c``.  If ``c`` is not assigned a grade by any voter, return None."""

        grades_for_c = [-1 * self.grade_order.index(g(c)) for g in self.grade_functions if g.has_grade(c)] if self.use_grade_order else [g(c) for g in self.grade_functions if g.has_grade(c)]

        sorted_grades_for_c = sorted(grades_for_c)
        num_grades = len(sorted_grades_for_c)
        median_idx = num_grades // 2
        if num_grades % 2 == 0:
            median_grades = sorted_grades_for_c[median_idx - 1:median_idx + 1]
        else:
            median_grades = [sorted_grades_for_c[median_idx]]

        if use_lower:
            return (self.grade_order[-1 * median_grades[0]] if self.use_grade_order else median_grades[0]) if self.has_grade(c) else None
        elif use_average:
            return (np.average([self.grade_order[-1 * m] for m in median_grades]) if self.use_grade_order else np.average(median_grades)) if self.has_grade(c) else None
        else:
            return ([self.grade_order[-1 * m] for m in median_grades] if self.use_grade_order else median_grades) if self.has_grade(c) else None

    def sum_grade_function(self):
        """Return the sum grade function of the profile."""

        assert self.can_sum_grades, "The grades in the profile cannot be summed."

        return _Mapping(
            {
                c: self.sum(c)
                for c in self.candidates if self.has_grade(c)
            },
            domain=self.candidates, 
            item_map=self.cmap, 
            compare_function=self.compare_function
        )
    def avg_grade_function(self):
        """Return the average grade function of the profile."""

        assert self.can_sum_grades, "The grades in the profile cannot be summed."

        return _Mapping(
            {
                c: self.avg(c)
                for c in self.candidates if self.has_grade(c)
            },
            domain=self.candidates, 
            item_map=self.cmap, 
            compare_function=self.compare_function
        )


    def proportion_with_grade(self, cand, grade):
        """
        Return the proportion of voters that assign a ``grade`` to ``cand`` .
        """

        assert cand in self.candidates, f"{cand} is not a candidate in the profile."
        assert grade in self.grades, f"{grade} is not a grade in the profile."

        num_with_higher_grade = 0
        for g,num in zip(*self.grades_counts):
            if self.compare_function(g(cand),grade) == 0: 
                num_with_higher_grade += num
        return num_with_higher_grade / self.num_voters


    def proportion_with_higher_grade(self, cand, grade):
        """
        Return the proportion of voters that assign a strictly higher grade to ``cand`` than ``grade``.
        """
        
        assert cand in self.candidates, f"{cand} is not a candidate in the profile."
        assert grade in self.grades, f"{grade} is not a grade in the profile."

        num_with_higher_grade = 0
        for g,num in zip(*self.grades_counts):
            if self.compare_function(g(cand),grade) == 1: 
                num_with_higher_grade += num
        return num_with_higher_grade / self.num_voters


    def proportion_with_lower_grade(self, cand, grade):
        """
        Return the proportion of voters that assign a strictly lower grade to ``cand`` than ``grade``.
        """

        assert cand in self.candidates, f"{cand} is not a candidate in the profile."
        assert grade in self.grades, f"{grade} is not a grade in the profile."


        num_with_lower_grade = 0
        for g,num in zip(*self.grades_counts):
            if self.compare_function(g(cand),grade) == -1: 
                num_with_lower_grade += num
        return num_with_lower_grade / self.num_voters

    def approval_scores(self): 
        """
        Return a dictionary representing the approval scores of the candidates in the profile.
        """

        assert self.can_sum_grades, "The grades in the profile cannot be summed."
        assert sorted(self.grades) == [0,1], "The grades in the profile must be 0 and 1."

        return {c: self.sum(c) for c in self.candidates}
    
    def to_ranking_profile(self): 
        """Return a ranking profile (a :class:ProfileWithTies) corresponding to the profile."""

        return ProfileWithTies(
            [g.ranking() for g in self._grades],
            rcounts = self.gcounts,
            candidates = self.candidates, 
            cmap = self.cmap
        )
    
    def write(self):
        """Write the profile to a string."""

        gprof_str = f"{len(self.candidates)};{self.num_voters};{self.grades}"
        for g in self.grade_functions: 
            g_str = ''
            for c in g.graded_candidates: 
                g_str += f"{c}:{g(c)},"
            gprof_str += f";{g_str[0:-1]}"
        return str(gprof_str)

    @classmethod
    def from_string(cls, gprof_str): 
        """
        Returns a profile of utilities described by ``gprof_str``.

        ``gprof_str`` must be in the format produced by the :meth:`pref_voting.GradeProfile.write` function.
        """
        gprof_data = gprof_str.split(";")

        num_alternatives,num_voters,grades = int(gprof_data[0]),int(gprof_data[1]),gprof_data[2:]

        grade_maps = [{int(cg.split(":")[0]): float(cg.split(":")[1]) for cg in gs.split(",")} if gs != '' else {} for gs in grades]

        if len(grade_maps) != num_voters: 
            raise Exception("Number of voters does not match the number of utilities.")
        
        return cls(grade_maps, candidates=range(num_alternatives))

    def display(self, show_totals=False, average_median_ties=False):
        """Display a grade profile as an ascii table (using tabulate). If ``show_totals`` is true then the sum, min, and max of the grades are displayed.

        """
                        
        if show_totals:
            sum_grade_fnc = self.sum_grade_function()
            headers = [''] + self.gcounts + ["Sum", "Median"]
            tbl = [[self.cmap[c]] + [self.gmap[g(c)] if g.has_grade(c) else "" for g in self._grades] + [sum_grade_fnc(c), self.median(c)] for c in self.candidates]
 
        else: 
            headers = [''] + self.gcounts
            tbl = [[self.cmap[c]] + [self.gmap[g(c)] if g.has_grade(c) else "" for g in self._grades] for c in self.candidates]
        print(tabulate(tbl, headers=headers))

    def visualize(self):
        """Visualize a grade profile as a stacked bar plot."""
        data_for_df = {
            'Candidate': [],
            'Grade': [],
            'Proportion': []
        }

        for c in self.candidates: 
            for g in [None] + self.grades:
                data_for_df['Candidate'].append(self.cmap[c])
                data_for_df['Grade'].append(self.gmap[g] if g is not None else "None")
                data_for_df['Proportion'].append(self.proportion(c, g))
        df = pd.DataFrame(data_for_df)

        # Pivot the DataFrame to organize it for the stacked bar plot
        df_pivot = df.pivot(index='Candidate', columns='Grade', values='Proportion')

        df_pivot
        # # Create the stacked bar plot
        ax = df_pivot.plot(kind='barh', stacked=True, figsize=(10, 6), rot=0)
        ax.set_ylabel('Candidate')
        ax.set_xlabel('Proportion')
        for spine in ['top', 'right']:
            ax.spines[spine].set_visible(False)
        # Show the plot

        ax.legend(loc='upper center', bbox_to_anchor=(0.5, 1.1), ncol=len(self.grades) + 1, title="Grades")

        plt.show()



'''
    File: utility_methods.py
    Author: Wes Holliday (wesholliday@berkeley.edu) and Eric Pacuit (epacuit@umd.edu)
    Date: May 26, 2023
    
    Implementations of utility methods. 
'''
from pref_voting.voting_method import vm
from pref_voting.social_welfare_function import swf
from pref_voting.rankings import  Ranking
import numpy as np  

@swf(name="Sum Utilitarian")
def sum_utilitarian_ranking(uprof, curr_cands = None): 
    """Rank the alternatives according to the sum of the utilities. 
    Args:
        uprof (Profile): A Profile object.
        curr_cands (list): A list of candidates to restrict the ranking to. If ``None``, then the ranking is over the entire domain of the profile.
    Returns:
        Ranking: A ranking of the candidates in ``curr_cands`` according to sum of the utilities.
    """

    curr_cands = curr_cands if curr_cands is not None else uprof.domain

    sums = {c:uprof.util_sum(c) for c in curr_cands}
    sorted_sums = sorted(list(set(sums.values())), reverse=True)
    return Ranking({x: uidx+1 for uidx, u in enumerate(sorted_sums) 
                    for x in curr_cands if sums[x] == u})

@vm(name="Sum Utilitarian")
def sum_utilitarian(uprof, curr_cands=None): 
    """
    Return the winning set of candidates according to sum of the utilities.
    """
    return sorted(sum_utilitarian_ranking(uprof, curr_cands=curr_cands).first())


@swf(name="Relative Utilitarian")
def relative_utilitarian_ranking(uprof, curr_cands=None): 
    """
    Rank the alternatives according to sum of the normalized utilities. 

    Args:
        uprof (Profile): A Profile object.
        curr_cands (list): A list of candidates to restrict the ranking to. If ``None``, then the ranking is over the entire domain of the profile.
    Returns:
        Ranking: A ranking of the candidates in ``curr_cands`` according to sum of the normalized utilities.

    .. note::
        Before restricting to curr_cands, we normalize with respect to *all* alternatives in the domain. 
    """

    curr_cands = curr_cands if curr_cands is not None else uprof.domain

    rel_utils = [u.normalize_by_range() for u in uprof.utilities]
    sums = {c:np.sum([u(c) for u in rel_utils if u(c) is not None]) for c in curr_cands}
    sorted_sums = sorted(list(set(sums.values())), reverse=True)
    return Ranking({x: uidx+1 for uidx, u in enumerate(sorted_sums) 
                    for x in curr_cands if sums[x] == u})

@vm(name="Relative Utilitarian")
def relative_utilitarian(uprof, curr_cands=None): 
    """
    Return the winning set of candidates according to sum of the normalized utilities.
    """
    return sorted(relative_utilitarian_ranking(uprof, curr_cands=curr_cands).first())


@swf(name="Maximin")
def maximin_ranking(uprof, curr_cands=None): 
    """
    Rank the alternatives according to the minimum utility.

    Args:
        uprof (Profile): A Profile object.
        curr_cands (list): A list of candidates to restrict the ranking to. If ``None``, then the ranking is over the entire domain of the profile.
    Returns:    
        Ranking: A ranking of the candidates in ``curr_cands`` according to minimum utility.
    """
    
    curr_cands = curr_cands if curr_cands is not None else uprof.domain

    min_utils = {x:uprof.util_min(x) for x in curr_cands}
    sorted_min_utils = sorted(list(set(min_utils.values())), reverse=True)
    return Ranking({x: midx+1 
                    for midx, m in enumerate(sorted_min_utils) 
                    for x in min_utils.keys() if min_utils[x] == m})

@vm(name="Maximin")
def maximin(uprof, curr_cands=None):
    """
    Return the winning set of candidates according to minimum utility.
    """
    return sorted(maximin_ranking(uprof, curr_cands=curr_cands).first())


@swf(name="Lexicographic Maximin")
def lexicographic_maximin_ranking(uprof, curr_cands=None): 
    """
    Rank the alternatives according to the lexicographic maximin ranking. The lexicographic maximin ranking is the ranking that ranks alternatives according to the minimum utility, and then breaks ties by ranking alternatives according to the second minimum utility, and so on.

    Args:
        uprof (Profile): A Profile object.
        curr_cands (list): A list of candidates to restrict the ranking to. If ``None``, then the ranking is over the entire domain of the profile.
    Returns:
        Ranking: A ranking of the candidates in ``curr_cands`` according to lexicographic maximin ranking.
    """

    curr_cands = curr_cands if curr_cands is not None else uprof.domain

    utils = {x: tuple(sorted([u(x) for u in uprof.utilities if u(x) is not None])) for x in curr_cands}
    assert len(list(set([len(us) for us in utils.values()]))) == 1, "Not all the items have the same number of utilities."
    sorted_utils = sorted(list(set(utils.values())), reverse=True)
    return Ranking({x: idx+1 
                    for idx, us in enumerate(sorted_utils) 
                    for x in utils.keys() if utils[x] == us})

@vm(name="Lexicographic Maximin")
def lexicographic_maximin(uprof, curr_cands=None):
    """
    Return the winning set of candidates according to lexicographic maximin ranking.
    """
    return sorted(lexicographic_maximin_ranking(uprof, curr_cands=curr_cands).first())


@swf(name="Nash")
def nash_ranking(uprof, sq=None, curr_cands=None): 
    """
    Rank the alternatives according to the Nash product ranking. Given the status quo ``sq``, the Nash product ranking ranks alternatives according to the product of the utilities of the alternatives minus the utility of the status quo.
    
    Args:
        uprof (Profile): A Profile object.
        sq (Candidate): The status quo. If ``None``, then the status quo is the first candidate in the domain of the profile.
        curr_cands (list): A list of candidates to restrict the ranking to. If ``None``, then the ranking is over the entire domain of the profile.
    Returns:
        Ranking: A ranking of the candidates in ``curr_cands`` according to Nash product ranking.
    
    """
    
    assert sq is None or sq in uprof.domain, f"The status quo {sq} must be in the domain of the profile."
    
    curr_cands = curr_cands if curr_cands is not None else uprof.domain

    sq = curr_cands[0] if sq is None else sq

    items_to_rank = list(set([x for x in curr_cands if all([u(x) > u(sq) 
                                                            for u in uprof.utilities])] + [sq]))
    
    nash_utils = {x: np.prod([u(x) - u(sq) for u in uprof.utilities]) 
                  for x in items_to_rank}
    sorted_nash_utils = sorted(list(set(nash_utils.values())), reverse=True)
    
    return Ranking({x: nidx+1 
                    for nidx, n in enumerate(sorted_nash_utils) 
                    for x in nash_utils.keys() if nash_utils[x] == n})

@vm(name="Nash")
def nash(uprof, sq=None, curr_cands=None):
    """
    Return the winning set of candidates according to Nash product ranking.
    """
    return sorted(nash_ranking(uprof, sq=sq, curr_cands=curr_cands).first())

utilitarian_vms = [
    sum_utilitarian,
    relative_utilitarian,
    maximin,
    lexicographic_maximin,
    nash
]

utilitarian_swfs = [
    sum_utilitarian_ranking,
    relative_utilitarian_ranking,
    maximin_ranking,
    lexicographic_maximin_ranking,
   # nash_ranking
]


from pref_voting.profiles import Profile
from pref_voting.profiles_with_ties import ProfileWithTies
from pref_voting.weighted_majority_graphs import MajorityGraph
from pref_voting.rankings import Ranking
from pref_voting.social_welfare_function import *
from pref_voting.voting_method import *
from itertools import combinations, chain
import random

import networkx as nx

def get_mg(edata, curr_cands = None): 
    
    if curr_cands == None: 
        if type(edata) == Profile or type(edata) == ProfileWithTies: 
            mg = MajorityGraph.from_profile(edata).mg
        else:
            mg = edata.mg
    else: 
        if type(edata) == Profile or type(edata) == ProfileWithTies:  
            mg = nx.DiGraph()
            mg.add_nodes_from(curr_cands)
            mg.add_edges_from([(c1,c2) for c1 in curr_cands for c2 in curr_cands if edata.majority_prefers(c1, c2)])
        else:
            mg = edata.mg.copy()
            mg.remove_nodes_from([c for c in edata.candidates if c not in curr_cands])
    return mg


def get_weak_mg(edata, curr_cands = None): 
    
    if curr_cands == None: 
        if type(edata) == Profile or type(edata) == ProfileWithTies: 
            wmg = MajorityGraph.from_profile(edata).mg
        else:
            wmg = edata.mg
        wmg.add_edges_from([(c1, c2) for c1 in edata.candidates for c2 in edata.candidates if c1 != c2 and edata.is_tied(c1, c2)])
    else: 
        if type(edata) == Profile or type(edata) == ProfileWithTies:  
            wmg = nx.DiGraph()
            wmg.add_nodes_from(curr_cands)
            wmg.add_edges_from([(c1,c2) for c1 in curr_cands for c2 in curr_cands if c1 != c2 and (edata.majority_prefers(c1, c2) or edata.is_tied(c1, c2))])
        else:
            wmg = edata.mg.copy()
            wmg.remove_nodes_from([c for c in edata.candidates if c not in curr_cands])
            wmg.add_edges_from([(c1, c2) for c1 in curr_cands for c2 in curr_cands if c1 != c2 and edata.is_tied(c1, c2)])
    return wmg


def swf_from_vm(vm, tie_breaker = None):
    """
    Given a voting method, returns a social welfare function that uses the voting method to rank the candidates (winners are ranked first; then they are excluded from curr_cands and the new winners are ranked second; etc.).

    Args:
        vm (function): A voting method.
        tie_breaker (str): The tie-breaking method to use. Options are "alphabetic", "random", and None. Default is None.

    Returns:
        function: A social welfare function that uses the voting method to rank the candidates.
    """
    
    def f(prof, curr_cands = None):

        cands = prof.candidates if curr_cands == None else curr_cands

        ranked_cands = list()
        ranking_dict = dict()

        n=0

        while n < len(cands):

            if len(ranked_cands) == len(cands):
                break

            ws = vm(prof, curr_cands = [c for c in cands if c not in ranked_cands])
            ranked_cands = ranked_cands + ws

            if tie_breaker is None:
                for c in ws:
                    ranking_dict[c] = n
                n += 1

            if tie_breaker == "alphabetic":
                sorted_ws = sorted(ws)
                for c in sorted_ws:
                    ranking_dict[c] = n
                    n += 1

            if tie_breaker == "random":
                random.shuffle(ws)
                for c in ws:
                    ranking_dict[c] = n
                    n += 1            

        return Ranking(ranking_dict)
        
    return SocialWelfareFunction(f, name = f"SWF from {vm.name}")


def vm_from_swf(swf):
    """
    Given a social welfare function, returns a voting method that selects all the candidates ranked first according to the swf.

    Args:
        swf (function): A social welfare function.

    Returns:
        function: A voting method that uses the swf to find the winning set.
    """
    
    def f(edata, curr_cands = None):
        return sorted(swf(edata, curr_cands = curr_cands).first())
        
    return VotingMethod(f, name = f"VM from {swf.name}")


def create_election(ranking_list, 
                    rcounts = None,
                    using_extended_strict_preference=None, 
                    candidates=None):
    """Creates an election from a list of rankings.
    
    Args:
        ranking_list (list): A list of rankings, which may be a list of tuples of candidates, a list of dictionaries or a list of Ranking objects.
        using_extended_strict_preference (bool, optional): Whether to use extended strict preference after creating a ProfileWithTies. Defaults to None.
        candidates (list, optional): A list of candidates.  Only used for creating a ProfileWithTies. Defaults to None (by default the candidates are all the candidates that are ranked by at least on voter).
    
    Returns:
        Profile or ProfileWithTies: The election profile.
    """

    if len(ranking_list) > 0 and (type(ranking_list[0]) == tuple or type(ranking_list[0]) == list):
        return Profile(ranking_list, rcounts=rcounts)
    elif len(ranking_list) > 0 and (type(ranking_list[0]) == dict or type(ranking_list[0]) == Ranking):
        if candidates is not None:
            prof = ProfileWithTies(ranking_list, candidates=candidates, rcounts=rcounts)
        else:
            prof = ProfileWithTies(ranking_list, rcounts=rcounts)       
        if using_extended_strict_preference:
            prof.use_extended_strict_preference()
        return prof
    else: # ranking_list is empty
        print("Warning: list of rankings is empty.")
        return Profile(ranking_list)
    

class SPO(object):
    """A strict partial order class due to Jobst Heitzig.
    
    The strict partial order P as a binary relation is encoded as a 2d numpy array.  The predecessors and successors of each object are precomputed.  The add method adds a new pair to the relation and computes the transitive closure.

    Args:
        n (int): The number of objects.
    
    """

    n = None
    """The number of objects"""
    objects = None
    """The list of objects"""
    P = None
    """The strict partial ordering P as a binary relation encoded as a 2d numpy array"""
    preds = None
    """The list of predecessors of each object"""
    succs = None
    """The list of successors of each object"""

    def __init__(self, n):
        self.n = n
        self.objects = list(range(n))
        self.P = np.zeros((n, n), dtype=bool)
        self.preds = [[] for _ in range(n)]
        self.succs = [[] for _ in range(n)]

    def add(self, a, b):
        """add a P b and all transitive consequences"""
        if not self.P[a][b]:
            self.P[a][b] = True
            self.preds[b].append(a)
            self.succs[a].append(b)
            for c in self.preds[a]:
                self._register(c, b)
                for d in self.succs[b]:
                    self._register(c, d)
            for d in self.succs[b]:
                self._register(a, d)

    def initial_elements(self):
        """return the initial elements of P (those without predecessors))"""
        return [i for i in self.objects if len(self.preds[i]) == 0]    

    def _register(self, a, b):
        """register that a P b, without forming the transitive closure"""
        if not self.P[a][b]:
            self.P[a][b] = True
            self.preds[b].append(a)
            self.succs[a].append(b)

    def to_numpy(self):
        """Return the partial order matrix P as a numpy array."""
        return self.P
    
    def to_networkx(self, cmap=None):
        """Convert the SPO to a networkx DiGraph.
        
        Args:
            cmap (dict): A dictionary mapping each number to a candidate name. If None, the identity map is used.
        
        Returns:
            nx.DiGraph: The resulting directed graph with nodes labeled according to cmap if provided.
        """
        G = nx.DiGraph()

        # Determine node labels based on cmap
        if cmap is not None:
            node_labels = {i: cmap[i] for i in self.objects}
        else:
            node_labels = {i: i for i in self.objects}

        # Add nodes with labels
        G.add_nodes_from(node_labels.values())

        # Add edges based on the partial order matrix
        for a in range(self.n):
            for b in range(self.n):
                if self.P[a][b]:
                    G.add_edge(node_labels[a], node_labels[b])

        return G
    
    def to_list(self, cmap=None):
        """If the SPO is a linear order, return a list representing the order.
        
        The list will contain candidate names based on the cmap if provided; otherwise, it will
        contain the numbers. 
        
        Returns None if the SPO is not a linear order.

        Args:
            cmap (dict): A dictionary mapping each number to a candidate name. If None, the identity map is used.
        """
        # Check if the SPO is a strict linear order
        for i in range(self.n):
            for j in range(i + 1, self.n):
                if not (self.P[i][j] or self.P[j][i]):
                    return None  # i and j are not comparable

        # Create the linear order list by topologically sorting the nodes
        linear_order = []
        visited = [False] * self.n

        def visit(node):
            if not visited[node]:
                visited[node] = True
                for successor in self.succs[node]:
                    visit(successor)
                linear_order.append(node)

        for node in self.objects:
            visit(node)

        # Reverse to get the correct order
        linear_order = linear_order[::-1]

        # If cmap is provided, map the numbers to candidate names
        if cmap is not None:
            linear_order = [cmap[node] for node in linear_order]
        
        return linear_order

def weak_orders(A):
    """A generator for all weak orders on A"""
    if not A:
        yield {}
        return
    for k in range(1, len(A) + 1):
        for B in combinations(A, k):
            for order in weak_orders(set(A) - set(B)):
                new_order = {cand: rank + 1 for cand, rank in order.items()}
                yield {**new_order, **{cand: 0 for cand in B}}


def weak_compositions(n, k):
    """A generator for all weak compositions of n into k parts"""

    if k == 1:
        yield [n]
    else:
        for i in range(n + 1):
            for comp in weak_compositions(n - i, k - 1):
                yield [i] + comp

def compositions(n):
    """Generates all compositions of the integer n. Adapted from https://stackoverflow.com/questions/10244180/python-generating-integer-partitions."""

    a = [0 for i in range(n + 1)]
    k = 1
    a[0] = 0
    a[1] = n
    while k != 0:
        x = a[k - 1] + 1
        y = a[k] - 1
        k -= 1
        while 1 <= y:
            a[k] = x
            x = 1
            y -= x
            k += 1
        a[k] = x + y
        yield a[:k + 1]

def enumerate_compositions(int_list):
    """Given a list of integers, enumerate all the compositions of the integers."""

    first_int = int_list[0]

    if len(int_list) == 1:
        for composition in compositions(first_int):
            yield [composition]

    else:
        for composition in compositions(first_int):
            for comps in enumerate_compositions(int_list[1:]):
                yield [composition] + comps

def sublists(lst, length, x = None, partial_sublist = None): 
    """Generate all sublists of lst of a specified length."""
    
    x = length if x is None else x
    
    partial_sublist = list() if partial_sublist is None else partial_sublist
    
    if len(partial_sublist) == length: 
        yield partial_sublist
        
    for i,el in enumerate(lst):
        
        if i < x: 
            
            extended_partial_sublist = partial_sublist + [el]
            x += 1
            yield from sublists(lst[i+1::], length, x, extended_partial_sublist)

def convex_lexicographic_sublists(l):
    """Given a list l, return all convex sublists S such that S is already sorted lexicographically."""

    cl_sublists = []
    current_list = []

    for idx, p in enumerate(l):
        if current_list + [p] == sorted(current_list + [p]):
            current_list = current_list + [p]

            if idx == len(l)-1:
                cl_sublists.append(current_list)

        else:
            cl_sublists.append(current_list)
            current_list = [p]
            
            if idx == len(l) - 1:
                cl_sublists.append(current_list)

    return cl_sublists

def powerset(iterable):
    """powerset([1,2,3]) --> () (1,) (2,) (3,) (1,2) (1,3) (2,3) (1,2,3)"""
    s = list(iterable)
    return chain.from_iterable(combinations(s, r) for r in range(len(s)+1))